//---------------------------------------------------------------------------

#pragma hdrstop

#include <stdlib.h>
#include <ctype.h>
#include "eq.h"
#include "group.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

void Eq::Error(const char *msg)
{
  Application->MessageBox(msg, Application->Title.c_str(), MB_OK+MB_ICONSTOP);
  exit(1);
}

// returns pointer to new group
int Eq::AddGroup()
{
  int i;

  if (_groups == MAXPTR)
    Error(__FUNC__ ": MAXPTR hit");
  i = _groups++;

  return i;
}

// returns pointer to added literal
int Eq::AddLiteral(const int value)
{
  int i;

  if (_literals == MAXPTR)
    Error(__FUNC__ ": MAXPTR hit");
  i = _literals++;
  _literal[i] = value;

  return i;
}

// returns pointer to added variable
int Eq::AddVariable(const char *name)
{
  int i;

  for(i = 0; i < _variables; i++)
    if (!strcmp(_variable[i].name, name))
      return i;
  if (_variables == MAXPTR)
    Error(__FUNC__ ": MAXPTR hit");
  i = _variables++;
  strcpy(_variable[i].name, name);

  return i;
}

// returns the priority of the passed group type
int Eq::Priority(const int type)
{
  switch (type)
  {
    case ADD:   return 0;
    case SUB:   return 0;
    case MUL:   return 1;
    case DIV:   return 1;
    case EXP:   return 2;
    case SQRT:  return 3;
    default:
      Error(__FUNC__ ": invalid type passed");
      return -1;
  }
}

int Eq::GetType(const char c)
{
  switch(c)
  {
    case '+': return ADD;
    case '-': return SUB;
    case '*': return MUL;
    case '/': return DIV;
    case '^': return EXP;
    default:
      Error(__FUNC__ ": unrecognized character in input");
      return -1;
  }
}

int Eq::GetTok(const char *str, int &p, char *buf, int &ptr, char &c)
{
  int i;

  if (!str[p])
    Error(__FUNC__ ": input terminated unexpectedly");
  i = 0;
  if (isdigit(str[p]))
  {
    while (isdigit(str[p]))
      buf[i++] = str[p++];
    buf[i] = 0;
    i = atoi(buf);
    ptr = AddLiteral(i);
    return T_LITERAL;
  }
  else if (isalpha(str[p]))
  {
    while (isalnum(str[p]))
    {
      if (i == MAXNAME)
        Error(__FUNC__ ": name exceeded MAXNAME characters");
      buf[i++] = str[p++];
    }
    buf[i] = 0;
    ptr = AddVariable(buf);
    return T_NAME;
  }
  else
  {
    c = str[p++];
    return T_OPERATOR;
  }
}

Eq::Member Eq::DoParse(const char *str, int &p,
                        const int l_prior, const Member l_member)
{
  int group,
      ptr,
      prior;
  char buf[256],
       c;
  int i,
      type;
  Member member,
         gmember;

  type = GetTok(str, p, buf, ptr, c);
  if (type == T_OPERATOR)
  {
    if (c == '(')
      member = DoParse(str, p, -1, gmember);
    else
      Error(__FUNC__ ": operand expected");
  }
  else if (type == T_LITERAL)
  {
    member.type = LITERAL;
    member.ptr = ptr;
  }
  else
  {
    member.type = VARIABLE;
    member.ptr = ptr;
  }
  if (!str[p])
    return member;
  type = GetTok(str, p, buf, ptr, c);
  if (type != T_OPERATOR)
    Error(__FUNC__ ": operator expected");
  if (c == ')')
    return member;
  type = GetType(c);
  prior = Priority(type);
  group = AddGroup();
  _group[group].type = type;
  gmember.type = GROUP;
  gmember.ptr = group;
  _group[group].right = DoParse(str, p, prior, gmember);
  if (l_prior > prior)
  {
    _group[group].left = l_member;
    return member;
  }
  else
  {
    _group[group].left = member;
    member.type = GROUP;
    member.ptr = group;
    return member;
  }
}

Eq::Eq()
{
  _symbols = 0;
}

void Eq::Parse(const char *str)
{
  int p;
  Member m;

  _groups = 0;
  _literals = 0;
  _variables = 0;

  p = 0;
  DoParse(str, p, -1, m);
  DumpGroups(str);
}

char GetGMemberType(int i)
{
  switch (i)
  {
    case GROUP:     return 'G';
    case LITERAL:   return 'L';
    case VARIABLE:  return 'V';
    case SYMBOL:    return 'S';
    default:        return '?';
  }
}

void Eq::DumpGroups(const char *str)
{
  int i;
  char c, buf[32];

  GroupWnd->Clear();
  GroupWnd->AddLine("Equation: %s", str);
  GroupWnd->AddLine("");

  GroupWnd->AddLine("Groups:");
  for(i = 0; i < _groups; i++)
  {
    switch (_group[i].type)
    {
      case ADD:   strcpy(buf, "add");   break;
      case SUB:   strcpy(buf, "sub");   break;
      case MUL:   strcpy(buf, "mul");   break;
      case DIV:   strcpy(buf, "div");   break;
      case EXP:   strcpy(buf, "exp");   break;
      default:    strcpy(buf, "???");
    }
    GroupWnd->AddLine("%3i: %6s: %c%-3i %c%-3i", i, buf,
                GetGMemberType(_group[i].left.type), _group[i].left.ptr,
                GetGMemberType(_group[i].right.type), _group[i].right.ptr);
  }
  GroupWnd->AddLine("");

  GroupWnd->AddLine("Literals:");
  for(i = 0; i < _literals; i++)
    GroupWnd->AddLine("%3i: %i", i, _literal[i]);
  GroupWnd->AddLine("");

  GroupWnd->AddLine("Variables:");
  for(i = 0; i < _variables; i++)
    GroupWnd->AddLine("%3i: %s", i, _variable[i].name);
  GroupWnd->AddLine("");

  GroupWnd->AddLine("Symbols:");
  for(i = 0; i < _symbols; i++)
    GroupWnd->AddLine("%3i: %8s = %i", i, _symbol[i].name, _symbol[i].val);
}

void Eq::PrettyPrint(Graphics::TBitmap *b)
{
  TFont *font;
  b->Canvas->Brush->Color = clWhite;
  b->Canvas->FillRect(Rect(0, 0, b->Width, b->Height));
  font = new TFont;
  font->Color = clBlack;
  font->Name = "Terminal";
  font->Height = 8;
  b->Canvas->Font = font;

//  b->Canvas->Brush->Color = clBlack;
//  b->Canvas->FillRect(Rect(16, 0, 24, 8));

  b->Canvas->Brush->Color = clWhite;
  b->Canvas->TextOut(0, 0, );

  delete font;
}


